/*************************************************************************/
/*                                                                       */
/*   Program Name :  m5tok5Rnp                                           */
/*                                                                       */
/*     M5tok5Rnp is a copy of m5tok5R on 2004-09-23 (Ver 1.30)           */
/*       both parity and non-parity modes are supported                  */
/*                                                                       */
/*     Enviroment variables                                              */
/*           M5DIR      ---   default directory for Mark5 data           */
/*           M5VEX      ---   default directory for Mark5 vex file       */
/*                                                                       */
/*      Version:  1.30  2004-09-23                                       */
/*                1.31  2011-02-09                                      */
/*                                                                       */
/*  copyright (c) 2003-2011 T.Kondo/NICT All Right Reserved              */
/*                                                                       */
/*************************************************************************/
/*

     2011-02-09  -- Ver 1.31  scanf("%s",&XX) was changed to scanf("%s",XX) when XX is 
                              declared as like "char XX[20]" to avoid warning at Ubuntu 10.10
                              scanf(...) was changed to like i=scanf(...) to avoid warning
                              at Ubuntu 10.10

  Mark-5 to K5 converter  (general purpose)

          2004-09-23  copied from m5tok5R

    m5tok5Rnp mk5_filename [options]
    
     where   mk5filename -- Mark 5 data file name
              options (any order)
               -c channel    -- pickup channel (1-16) for 1ch conversion mode
                                 (default all group and 4ch mode)
               -g group      -- pickup group (1-4) for 4 ch conversion mode
                                 (omitted when -c option is set)
                                 (default all group)
               -i info_file  -- infomation file name that contains
                                  track vs channel table
                                  bit_psition vs channel table
                                  group# vs channel table
                                  (default "m5tok5info.txt")
                                  "-i -info_file"  means info_file named info_file creation
                                  "-i make" means default info_file creation
               -o k5name     -- K5 file name to be created
                                 (default : see below)
               -d k5dir      -- directory for k5 file out
                                 (default  : mark5 directory)
               -s strat_sec  -- offset time from data head (sec)
                                 (default :  0 )
               -p period     -- data period to convert (sec)
                                 (default : all)
               -r            -- reverse track order when fanout =2 or more
               -v vex_file   -- VEX file name to be read when "-i make" is set
               -sid stat_id  -- station ID when "-i make" is set
               -scan scan#   -- scan# for MODE get when "-i make" is set
                                 (default : 1)\n");
               -subp subpass -- set subpass at information file 
                                 creation mode (default is "A") 
               -odd          -- 32bit (4byte) shift in reading Mark5 data
                                 (this works only for 64 track mode)

   
  Naming rule for K5 file created (MK5 : original mark5 name)
         4ch mode
            MK5.k5a ---- for group#1 in info_file (usually CH 01-04)
            MK5.k5b ---- for group#2 in info_file (usually CH 05-08)
            MK5.k5c ---- for group#3 in info_file (usually CH 09-12)
            MK5.k5d ---- for group#4 in info_file (usually CH 13-16)
         1ch mode
            MK5.k5-NN ---- where NN is channel number 01-16

  Enviroment variables
          M5DIR    ---   default path to Mark5 data
          M5VEX    ---   default path to Mark5 vex file

  Note:  When compiling on VC++, set stack size at least 4000000(4MB).
         otherwise results in runtime error.

*** sample run parameters ****

1. make a information file 
 case 1:  make a default information file (m5tok5info.txt)
  H:\IPVLBI\data\jiveexp3\d03l1_2.on -i make -v H:\IPVLBI\data\jiveexp3\d03l1.vex -sid On

 case 2:  make an information file named d03l1_on_info.txt
  H:\IPVLBI\data\jiveexp3\d03l1_2.on -i -H:\IPVLBI\data\jiveexp3\d03l1_on_info.txt -v H:\IPVLBI\data\jiveexp3\d03l1.vex -sid On

2. convert Mark5 file to K5 file
 case 1: using a default information file
  H:\IPVLBI\data\jiveexp3\d03l1_2.on -g 1

 case 2: using an information file named d03l1_info.txt
  H:\IPVLBI\data\jiveexp3\d03l1_2.on -g 1 -i H:\IPVLBI\data\jiveexp3\d03l1_on_info.txt

3. other examples for debugging purpose
  J:\rd0401\E085-1224 -i J:\rd0401\rd0401.Einfo.txt -g 1
  J:\jive\f03x1_5.ef -g 1 -i J:\jive\f03x1_info.txt
  H:\IPVLBI\data\evlbi3\288-1730 -i make -v H:\IPVLBI\data\evlbi3\2994.ovex
  H:\IPVLBI\data\evlbi3\288-1730 -g 1 -i H:\IPVLBI\data\evlbi3\evlbi3_info.txt
  J:\GSICHECK\K04242\242-0734 -g 1 -s 3 -i J:\GSICHECK\K04242\k04242info.txt
  J:\jive128Mbps\f04c2_7.hh -i  make -v J:\jive128Mbps\n04u1.vex

*/


#define PROGRAM_VERSION "m5tok5Rnp (Ver 1.30  2004-09-23)"

#include "libk5.h"      // common header for IP-VLBI utilities 

#include "libmk5tok5.h"  // library for Mark5 to K5 conversion

/* default conversion information file name set */
#define INFONAME_DFLT "m5tok5Rnpinfo.txt"   // information file default name
/* environment variable for Mark5 data default directory set */
#define M5DIR "M5DIR" // the environment variable for Mark5 data default directory
/* environment variable for Mark5 vex file default directory set */
#define M5VEX "M5VEX" // the environment variable for Mark5 vex file

/* default pass set */
#define DEFLT_PASS "A"   /* default subpass in $TRACKS in a VEX file (added on 2004-06-29) */

void param_get_m5tok5Rold(int argc, char *argv[],
               char *mk5name, char *k5name,
               char *outdir, char *info_file,
               int *channel, int *group, int *soffset, int *period,
               int *reverse, int *makemode,
               char *vex_file, char *sid, int *scan, char *dflt_pass)
{
    int n,i,len;
    char cbuf[80];
    //1st param should be mark5 name
    if (argc>=2){
       strcpy(mk5name,*(argv+1));
    } else {
       mk5name[0]='\0';
    }

    /* check Mark5 file name is acompanied by directory or not */
    get_true_path(mk5name, M5DIR, "./",cbuf);
    strcpy(mk5name,cbuf);

    // initialize parameters
    *channel=0;
    *group=0;
    *soffset=0;
    *period=0;
    *reverse=0;
    *makemode=0;
    info_file[0]='\0';
    k5name[0]='\0';
    outdir[0]='\0';
    vex_file[0]='\0';
    sid[0]='\0';
    *scan=1;
    strcpy(dflt_pass, DEFLT_PASS);

    n=3;
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));
        if(strncmp(cbuf,"-i",2)==0){  // info file
            n++;
            if(n > argc) break;
            strcpy(info_file,*(argv+n-1));
            if(info_file[0]=='-'){
                *makemode=1;
                len=strlen(info_file);
                for(i=0;i<len-1;i++) info_file[i]=info_file[i+1];
                info_file[len-1]='\0';
            } else if(strncmp(info_file,"make",4)==0){
                *makemode=1;
                info_file[0]='\0';
            }
        }
        if(strncmp(cbuf,"-o",2)==0){  // k5 name
            n++;
            if(n > argc) break;
            strcpy(k5name,*(argv+n-1));
        }
        if(strncmp(cbuf,"-d",2)==0){  // out directory
            n++;
            if(n > argc) break;
            strcpy(outdir,*(argv+n-1));
            /* env variable is substituted by real one */

            if((i=substitute_env_path_name(outdir, cbuf))>0){
                strcpy(outdir,cbuf);
            }

        }
        if(strncmp(cbuf,"-c",2)==0){  // channel define
            n++;
            if(n > argc) break;
            *channel = atoi(*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-g",2)==0){  // group define
            n++;
            if(n > argc) break;
            *group = atoi(*(argv + n-1 ));
        }
        //if(strncmp(cbuf,"-s",2)==0){  // start offset
        if(strlen(cbuf)==2 && strncmp(cbuf,"-s",2)==0){  // start offset
            n++;
            if(n > argc) break;
            *soffset = atoi(*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-p",2)==0){  // period
            n++;
            if(n > argc) break;
            *period = atoi(*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-r",2)==0){  // period
            *reverse=1;
        }
        if(strncmp(cbuf,"-v",2)==0){  // vex file
            n++;
            if(n > argc) break;
            strcpy(vex_file,*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-si",3)==0){  // station id
            n++;
            if(n > argc) break;
            strcpy(sid,*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-scan",5)==0){  // scan#
            n++;
            if(n > argc) break;
            *scan=atoi(*(argv + n-1 ));
            if(*scan < 1) *scan=1;
        }
        if(strncmp(cbuf,"-subp",5)==0){  // subpass
            n++;
            if(n > argc) break;
            strcpy(dflt_pass,*(argv + n-1 ));
        }
        n++;
    }
}

void param_get_m5tok5R(int argc, char *argv[],m5k5runp_t *runp)                       
{
    int n,i,len;
    char cbuf[60];
    //1st param should be mark5 name
    if (argc>=2){
       strcpy(runp->mk5name,*(argv+1));
    } else {
       runp->mk5name[0]='\0';
    }
     /* check Mark5 file name is acompanied by directory or not */
    get_true_path(runp->mk5name, M5DIR, "./",cbuf);
    strcpy(runp->mk5name,cbuf);

    // initialize parameters
    runp->channel=0;
    runp->group=0;
    runp->soffset=0;
    runp->period=0;
    runp->reverse=0;
    runp->makemode=0;
    runp->info_file[0]='\0';
    runp->k5name[0]='\0';
    runp->k5name_given[0]='\0';
    runp->outdir[0]='\0';

    runp->vex_file[0]='\0';
    runp->sid[0]='\0';
    runp->scan=1;
    strcpy(runp->dflt_pass, DEFLT_PASS);

    for(n=0; n<4; n++) runp->ibit[n]=0;
    runp->numch=1;
    runp->ptype=0;
    runp->odd=0;

    //check run parametees are old type or not
    if(argc>=3){
        strcpy(cbuf,*(argv+2) );
        if(cbuf[0] != '-') goto oldtype;  // old type parameters
    }

    n=3;
    while(n <=argc) {
        strcpy(cbuf,*(argv+n-1));
        if(strncmp(cbuf,"-i",2)==0){  // info file
            n++;
            if(n > argc) break;
            strcpy(runp->info_file,*(argv+n-1));
            if(runp->info_file[0]=='-'){
                runp->makemode=1;
                len=strlen(runp->info_file);
                for(i=0;i<len-1;i++) runp->info_file[i]=runp->info_file[i+1];
                runp->info_file[len-1]='\0';
            } else if(strncmp(runp->info_file,"make",4)==0){
                runp->makemode=1;
                runp->info_file[0]='\0';
            }
        }
        if(strlen(cbuf)==2 && strncmp(cbuf,"-o",2)==0){  // k5 name
            n++;
            if(n > argc) break;
            strcpy(runp->k5name_given,*(argv+n-1));
        }
        if(strncmp(cbuf,"-d",2)==0){  // out directory
            n++;
            if(n > argc) break;
            strcpy(runp->outdir,*(argv+n-1));
            /* env variable is substituted by real one */

            if((i=substitute_env_path_name(runp->outdir, cbuf))>0){
                strcpy(runp->outdir,cbuf);
            }

        }
        if(strncmp(cbuf,"-c",2)==0){  // channel define
            n++;
            if(n > argc) break;
            runp->channel = atoi(*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-g",2)==0){  // group define
            n++;
            if(n > argc) break;
            runp->group = atoi(*(argv + n-1 ));
        }
        if(strlen(cbuf)==2 && strncmp(cbuf,"-s",2)==0){  // start offset
            n++;
            if(n > argc) break;
            runp->soffset = atoi(*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-p",2)==0){  // period
            n++;
            if(n > argc) break;
            runp->period = atoi(*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-r",2)==0){  // 
            runp->reverse=1;
        }
        if(strncmp(cbuf,"-v",2)==0){  // vex file
            n++;
            if(n > argc) break;
            strcpy(runp->vex_file,*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-si",3)==0){  // station id
            n++;
            if(n > argc) break;
            strcpy(runp->sid,*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-scan",5)==0){  // scan#
            n++;
            if(n > argc) break;
            runp->scan=atoi(*(argv + n-1 ));
            if(runp->scan < 1) runp->scan=1;
        }
        if(strncmp(cbuf,"-subp",5)==0){  // subpass
            n++;
            if(n > argc) break;
            strcpy(runp->dflt_pass,*(argv + n-1 ));
        }
        if(strncmp(cbuf,"-odd",4)==0){  // odd mode set
            runp->odd=1;
        }
        n++;
    }
    goto exitend;
oldtype:
    // old type parameter analysis

    runp->ptype=1;
    if (argc>=3) strcpy(runp->k5name_given,*(argv+2));
    runp->numch=1;

    if (argc>=4) runp->ibit[0] = atoi(*(argv + 3 ));
    if (argc>=5){
        runp->numch=4;
        runp->ibit[1] = atoi(*(argv + 4 ));
    }
    if (argc>=6) runp->ibit[2] = atoi(*(argv + 5 ));
    if (argc>=7) runp->ibit[3] = atoi(*(argv + 6 ));

exitend:;
}

void print_usage(void)
{
    //printf("m5tok5Rnp  %s\n",M5TOK5RNP_VERSION);
    printf("%s\n",PROGRAM_VERSION);
    printf("Usage: m5tok5Rnp mk5_filename [options]\n");
    printf("\n");
    printf("where   mk5filename -- Mark 5 data file name\n");
    printf("       options (any order)\n");
    printf("        -c channel    -- pickup channel (1-16) for 1ch conversion mode\n");
    printf("                          (default all group and 4ch mode)\n");
    printf("        -g group      -- pickup group (1-4) for 4 ch conversion mode\n");
    printf("                          (omitted when -c option is set)\n");
    printf("                          (default all group)\n");
    printf("        -i info_file  -- infomation file name that contains\n");
    printf("                           track vs channel table\n");
    printf("                           bit_psition vs channel table\n");
    printf("                           group# vs channel table\n");
    printf("                           (default \"%s\")\n",INFONAME_DFLT);
    printf("                           \"-i -info_file\" means\n");
    printf("                             create infomation file named info_file\n");
    printf("                           \"-i make\" means create default info file\n");
    printf("        -o k5name     -- K5 file name to be created\n");
    printf("                          (default : see below)\n");
    printf("        -d k5dir      -- directory for k5 file out\n");
    printf("                          (default  : mark5 directory)\n");
    printf("        -s strat_sec  -- offset time from data head (sec)\n");
    printf("                          (default :  0 )\n");
    printf("        -p period     -- data period to convert (sec)\n");
    printf("                          (default : all)\n");
    printf("        -r            -- reverse track order when fanout =2 or more\n");
    printf("        -v vex_file   -- VEX file name to be read when \"information\n");
    printf("                              file creation mode\" is set\n");
    printf("        -sid stat_id  -- station ID when \"information file creation\n");
    printf("                              mode\" is set\n");
    printf("        -scan scan#   -- scan# for MODE get when \"information file\n");
    printf("                              creation mode\" is set (default : 1)\n");
    printf("        -subp subpass -- set subpass at information file\n");
    printf("                              creation mode (default is \"A\")\n");
    printf("        -odd          -- 32bit (4byte) shift in reading Mark5 data\n");
    printf("                          (this works only for 64 track mode)\n");
    printf("\n");
    printf("   Naming rule for K5 file created (MK5 : original mark5 name)\n");
    printf("      4ch mode\n");
    printf("        MK5.k5a ---- for group#1 in info_file (usually ch01-04)\n");
    printf("        MK5.k5b ---- for group#2 in info_file (usually ch05-08)\n");
    printf("        MK5.k5c ---- for group#3 in info_file (usually ch09-12)\n");
    printf("        MK5.k5d ---- for group#4 in info_file (usually ch13-16)\n");
    printf("      1ch mode\n");
    printf("        MK5.k5-NN ---- where NN is channel number 01-16\n");
    printf("   Enviroment variables\n");
    printf("        %s    ---   default path to Mark5 data ( %s )\n",
        M5DIR,getenv(M5DIR));
    printf("        %s    ---   default path to Mark5 vex file ( %s )\n",
        M5VEX,getenv(M5VEX));
    printf("\n");
}


int main(int argc, char *argv[])
{
       int i,m, k, n,tc;
       int obuf[1024];
       int nwrite=1024;  // obuf size to write
       int iwrite; 
       int iwork, iw, opos;
       int ipos, ipos2, offset;
	   int flag;
	   int pcount;
       int mk5data_mode; // Mk5 data mode
                         //  1: NRZM with Parity (NRZM->NRZL necessary)
                         //  2: NRZM without parity   (NRZM->NRZL necessary)
                         //  3: NRZL with parity
                         //  4: NRZL without parity
	   int iret; // 2011-02-09
       double fsampltrk;
       double telapsed;           // to measure processing time
	   time_t tstart, tfinish;    //           "
       char fname[50];
       char fpath[100];
       char program_version[60];
       char* dflt_iname=INFONAME_DFLT;   // information file default name
       FILE *fo;
       FILE *fin;
       IPBOARD_T ipb;
       mk5_data_t *mk5;    
       mk5tok5_info_t mk5info;
       gr_info_t gr;  
       m5k5runp_t runp;

       if(argc == 1){
          print_usage();
          exit(0);
       }
       if (argc>=2){
          if(strncmp(*(argv+1),"env",3)==0){
          printf("   Enviroment variables\n");
          printf("        %s ---   default directory for Mark5 data\n",M5DIR);
          printf("                      ( %s )\n",getenv(M5DIR));
          printf("        %s ---   default directory for Mark5 vex file\n",M5VEX);
          printf("                      ( %s )\n",getenv(M5VEX));
          printf("\n");
          exit(0);
          }
       }

       strcpy(program_version,PROGRAM_VERSION);

       /* Program Name Display */
       printf("****************************************************\n");
       printf("*    Mark-5 to K5/VSSP format converter            *\n");
       printf("*      %s            *\n",PROGRAM_VERSION);
       printf("*          Kashima Space Research Center/NICT      *\n");
       printf("****************************************************\n");
       printf("\n");

       /* get run parameters */
       param_get_m5tok5R(argc, argv, &runp); 

       /* default file information file name set if not given */
       if (runp.info_file[0]=='\0'){
          strcpy(runp.info_file,dflt_iname);
       }

       /* If out directory is not set, out directory is 
           set to the same path as that to mark5 file    */  
       if(runp.outdir[0]=='\0'){
          parse_dir_and_name(runp.mk5name, runp.outdir, fname); 
       }

       /* run condition monitor */
       printf(" ---------- RUN CONDITION -----------\n");
       printf("   mk5name   : %s\n",runp.mk5name);
       printf("   k5name    : %s\n",runp.k5name_given);
       printf("   outdir    : %s\n",runp.outdir);
       printf("   infofile  : %s\n",runp.info_file);
       printf("   channel   : %d\n",runp.channel);
       printf("   group     : %d\n",runp.group);
       printf("   soffset   : %d\n",runp.soffset);
       printf("   period    : %d\n",runp.period);
       printf("   odd       : %d\n",runp.odd);
       if(runp.vex_file[0] != '\0')
          printf("   vex_file  : %s\n",runp.vex_file);
       if(runp.sid[0] != '\0' && runp.sid[0] !='-')
          printf("   station_id: %s\n",runp.sid);
       if(runp.makemode==1){
          printf("   scan# for MODE get : %d\n",runp.scan);
          printf("   subpass to select  : %s\n",runp.dflt_pass);
       }
       printf(" -------------------------------------\n");
       printf("\n");

       /* Mark5 file existence check */
       if( (fin  = fopen( runp.mk5name, "rb" )) == NULL ){
           printf("Mark5 File %s   Open Error! \n",runp.mk5name);
		   goto exitend;
       }
       i=fclose(fin);

       /* information file create if selected so */

       if(runp.makemode ==1){  
           /*****************************************************/
           /* Entry of Information file creation mode operation */ 
           /*****************************************************/
           printf("mk5tok5R running under Information File create mode\n");
           printf("info file (%s) will be created (updated).\n",runp.info_file);
           if(runp.vex_file[0]=='\0'){  // interactive operation
              printf("\nEnter VEX file name -> ");
              //scanf("%s",&runp.vex_file);
              iret=scanf("%s",runp.vex_file);  // 2011-02-09
           } 

           /* get true path for vex file name  */
           get_true_path(runp.vex_file, M5VEX, "./",fpath);
           strcpy(runp.vex_file,fpath);
           printf("VEX file name --- %s\n",runp.vex_file);
           
           /* make information file by reading Vex file and by 
                analyzing Mark-5 file                          */
           make_mk5tok5_info_file(runp.info_file,runp.mk5name,
               runp.vex_file,runp.sid,runp.scan,
               program_version, runp.dflt_pass);
           printf("\nInformation file (%s) created!!\n",runp.info_file);
           printf("\n");
           exit(1);
           /***************************************************/
           /* End of Information file creation mode operation */ 
           /***************************************************/
       }

       /**************************************/
       /* Entry of Conversion mode operation */ 
       /**************************************/

       /* internal conversion information data set up */
       mk5tok5_info_setup(runp.info_file, &mk5info);

       /* monitor mk5info */
       monit_mk5info(&mk5info);  // debug    

       /* Mark5 data mode set */
       mk5data_mode=4; //   default   NRZL without parity
       if(mk5info.nrzm==1){
           if(mk5info.parity==1) {
               mk5data_mode=1;
           } else {
               mk5data_mode=2;
           }
       } else {
           if(mk5info.parity==1) {
               mk5data_mode=3;
           } else {
               mk5data_mode=4;
           }
       }
       //printf("mk5data_mode = %d\n",mk5data_mode);
       if(mk5data_mode==1) printf(" Mark5 data : NRZM encoded with parity is supposed\n");
       if(mk5data_mode==2) printf(" Mark5 data : NRZM encoded without parity is supposed\n");
       if(mk5data_mode==3) printf(" Mark5 data : with parity is supposed\n");
       if(mk5data_mode==4) printf(" Mark5 data : no parity-bit is supposed\n");


       tstart= time(NULL);    // to measure processing time

       tc=-1;
       offset=0;
       if(mk5info.vlba==1) offset=160;

       /* data validity check */
       k=0;
       for(n=0; n<mk5info.ntrack; n++) k+=mk5info.bit_valid[n] ;
       if(k == 0) {  // novalid track
           printf(" No valid track was found in the information file.\n");
           goto exitend;   
       }


       printf("\n Group# vs CH# table for conversion \n");
       printf("   Gr#           CH#\n");
       printf("  -----    ----------------\n");
       for(m=0; m<4; m++){
           printf("   %2d     ",m+1);
           for(k=0; k<4; k++){
               printf("%4d",mk5info.group_channel[m][k]);
           }
           printf("\n");
       }
       printf("  -----    ----------------\n\n");
	   
       gr.channel=runp.channel;        
       gr.group=runp.group;
       gr.reverse=runp.reverse;

   /********************************************/
   /* loop entry of Frequency group processing */ 
   /********************************************/

loop_g:     

       if(runp.channel==0 && runp.group==0) {
           gr.group++;
           if(mk5info.nch-4*(gr.group-1) <=0) goto exit_loop_g;
       }
       
       // setup processing channel number table etc.
       i=Grinfo_fm_Mk5info(&mk5info, &gr);
       if(i==-1) goto exit_loop_g;
       if((i==-2) && runp.group==0) goto loop_g;  // if all channel ==0 next group

       /* if odd mode is set and ntrack=64, edit bit infomation table */
       if(runp.odd==1 && mk5info.ntrack==64){
           printf(" bit-track table was edited\n from ");
           for(n=0; n<gr.tot_track; n++) printf(" %2d",gr.bit_trk[n]);
           printf("\n");
           for(n=0; n<gr.tot_track; n++){
               gr.bit_trk[n]+=32;
               gr.bit_trk[n]%=64 ;
           }
           printf(" to   ");
           for(n=0; n<gr.tot_track; n++) printf(" %2d",gr.bit_trk[n]);
           printf("\n");
       }
 
       /* conversion start message */
       if(runp.channel == 0){
           printf("*************************************************\n");
           printf("* Group# %2d : CH# %2d %2d %2d %2d conversion start! *\n",
               gr.group, 
               mk5info.group_channel[gr.group-1][0],
               mk5info.group_channel[gr.group-1][1],
               mk5info.group_channel[gr.group-1][2],
               mk5info.group_channel[gr.group-1][3]);
           printf("*************************************************\n");
       } else {
           printf("****************************\n");
           printf("* CH# %2d conversion start! *\n", gr.channel);
           printf("****************************\n");
       }

       /* actual bit position information for this processing */
       monit_bitpos_info(&mk5info, &gr);

       /* memory allocation Mk5 data type variables */
       mk5 = (mk5_data_t *)malloc(gr.tot_track*sizeof(mk5_data_t));   

 	   fsampltrk=mk5info.samplrate/mk5info.fanout;  // track sampling rate

       printf("Mark 5 Data File : %s  ..... Atamadashi start\n",runp.mk5name);
       printf("\n");
       printf("      bit pos#\n");
       printf("     K5   Mark-5     1st frame time         atamadashi time\n");
       printf("  -----------------------------------------------------------\n");

       //printf("  tot_track = %d\n",gr.tot_track);
	   for(n=0; n<gr.tot_track; n++){
           /* Data file open */
          if( (mk5[n].fin  = fopen( runp.mk5name, "rb" )) == NULL ){
              printf("Data File %s   Open Error! \n",runp.mk5name);
		      goto exitend;
		  }
          /* initialize mk5 structure info data */
          init_mk5_dat(&mk5[n],&mk5info);
		  mk5[n].bitpos=gr.bit_trk[n];   // bit position of this track

          /* atamadashi */

          i=read_1frame_r(&mk5[n]); 

	      if(i<=0) {
		     i=fclose(mk5[n].fin);
		     goto exitend;
		  }

          //for (i=0; i<5; i++){   //check           
          //  printf("%02d %08x ",n, rvsint(mk5[n].header[i]));
          //} 
	      //printf("\n");

          printf("  %5d  %4d      %d/%03d %02d:%02d:%02d.%03d -->",
               n,mk5[n].bitpos, 
               mk5[n].itime[0],mk5[n].itime[1],mk5[n].itime[2],
               mk5[n].itime[3],mk5[n].itime[4],mk5[n].itime[5]);

          /* atamadashi for just seconds start */

	      i=fast_atamadashi_r(&mk5[n], runp.soffset);  

	      if(i<=0) {
		     i=fclose(mk5[n].fin);
		     goto exitend;
		  }
/*
          printf("%5d   ",gr.bit_trk[n]);
          for (i=0; i<5; i++){   //check           
             printf("%08x ",rvsint(mk5[n].header[i]));
	 	  } 
	      printf("\n");
*/

          printf(" %d/%03d %02d:%02d:%02d.%03d\n",
               mk5[n].itime[0],mk5[n].itime[1],mk5[0].itime[2],mk5[n].itime[3],
		       mk5[n].itime[4],mk5[n].itime[5]);

       }
       printf("  -----------------------------------------------------------\n");
	   
       printf("Atamadashi finished \n");


       //debug
       //for(i=0; i<gr.tot_track; i++) printf(" %d",mk5[i].idata[159]);
       //printf("\n");


       /* output file create */
       if(runp.k5name_given[0]=='\0'){
           k5name_make(runp.mk5name, runp.outdir, runp.channel, gr.group, runp.k5name);
       } else {
           strcpy(runp.k5name,runp.k5name_given);
       }
       if( (fo  = fopen( runp.k5name, "wb" )) == NULL )
       {
         printf("File  %s  Open Error! \n",runp.k5name);
          goto exitend1;
	   } 

       /* K5 header data creation */
       ipb.adbit=mk5info.adbit;
       ipb.numch=gr.numch;
       ipb.sfreq=(int)(mk5info.samplrate/1000);
       printf("\nK5 data parameters : ");
       if(ipb.sfreq >= 1000){
           printf("%d MHz %d bit sampling  %d ch data\n",
               ipb.sfreq/1000, ipb.adbit, ipb.numch);
       } else {
           printf("%d kHz %d bit sampling  %d ch data\n",
               ipb.sfreq, ipb.adbit, ipb.numch);
       }
       printf("K5 file (%s) is being created...\n",runp.k5name);

       ipb.seconds=mk5[0].itime[2]*3600+mk5[0].itime[3]*60+mk5[0].itime[4];
       //printf("ipb.seconds %d\n",ipb.seconds);
       ip_hd_make_ini(&ipb);

       /* put first K5 header data into obuf */

       opos=0;
       obuf[opos]=conv4byte2int(ipb.hd);
       opos++;
       obuf[opos]=conv4byte2int(ipb.hd+4);
       opos++;


       /* conversion Mark-5 data to K5 by every frame */


       pcount=0;
	   ipos=offset;   // ipos : bit position counter in a frame
                      // Mark-IV: 0 -> 20000,   VLBA: 160-> 20160
	   ipos2=0;  // bit position counter in 4 bytes (0 -> 32)
	   iwork=0;
	   flag=0;
       while(1){
		  for(i=0; i<gr.tot_track; i++){
              iw=mk5[i].idata[ipos];
              
			  if(flag==0){
			     iwork=iwork >> 1;
			     iw=iw << 31;
                 iwork= iw | (iwork & 0x7FFFFFFF);
			  } else {
			     iwork=iwork << 1;
			     iw=iw & 0x1;
                 iwork= iw | (iwork & 0xFFFFFFFE);
			  }
              
			  ipos2++;
		  }
		  if(ipos2==32){
              //decodepatn(&iwork);  // no more use
			  obuf[opos]=iwork;
			  opos++;
              if(opos == nwrite){
                  iwrite=fwrite(obuf,sizeof(int),nwrite,fo);
                  opos=0;
                  //debug
                //  for(i=0; i<gr.tot_track; i++) printf(" %d",mk5[i].idata[159]);
                //  printf("\n");
                //  printf(" %08x\n",obuf[81]);
              }
			  ipos2=0;
			  iwork=0;
		  }
		  ipos++;
		  if(ipos==mk5info.bitsframe){  // 1 frame data were read 
			  ipos=offset;
        
			  for(n=0; n<gr.tot_track; n++){
                  if(mk5data_mode==1){
                     i=read_1frame_m_p(&mk5[n]); 
                  } else if(mk5data_mode==4){
                     i=read_1frame_l_np(&mk5[n]);
                  } else {
                     i=read_1frame_r(&mk5[n]); 
                  }
	              if(i<=0) {
		              goto exitloop;
				  }
			  }

		      if(mk5[0].itime[5]==0){ // this is just second
                  pcount++;
                  if(runp.period !=0 && pcount > runp.period) goto exitloop;

                  // debug
			      //for(n=0; n<gr.tot_track; n++){
                  //  for (i=0; i<5; i++){            
                  //      printf("%08x ",rvsint(mk5[n].header[i]));
                  //  } 
                  //  printf("\n");
                  //}

                  printf("  header time : %d/%03d %02d:%02d:%02d.%03d",
                    mk5[0].itime[0],mk5[0].itime[1],mk5[0].itime[2],mk5[0].itime[3],
			        mk5[0].itime[4],mk5[0].itime[5]);
                  if(gr.numch==4){
				      if(mk5[1].itime[5]+mk5[2].itime[5]+mk5[3].itime[5] !=0){
					      printf("   Warning!! time difference occurred!\n");

                          printf("%d/%03d %02d:%02d:%02d.%03d\n",
                             mk5[1].itime[0],mk5[1].itime[1],mk5[1].itime[2],
                             mk5[1].itime[3],mk5[1].itime[4],mk5[1].itime[5]);
                          printf("%d/%03d %02d:%02d:%02d.%03d\n",
                             mk5[2].itime[0],mk5[2].itime[1],mk5[2].itime[2],
                             mk5[2].itime[3],mk5[2].itime[4],mk5[2].itime[5]);
                          printf("%d/%03d %02d:%02d:%02d.%03d\n",
                             mk5[3].itime[0],mk5[3].itime[1],mk5[3].itime[2],
                             mk5[3].itime[3],mk5[3].itime[4],mk5[3].itime[5]);
                      } else {
                          //printf("   Good! (All CH Synchronized)\n");
                          printf("   Good!\n");
                      }
                  } else {
                      printf("\n");
                  }

				  /* put K5 header */
			      ipb.seconds+=1;
                  ip_hd_make_ini(&ipb);
                  obuf[opos]=conv4byte2int(ipb.hd);
                  opos++;
                  if(opos == nwrite){
                     iwrite=fwrite(obuf,sizeof(int),nwrite,fo);
                     opos=0;
				  }
                  obuf[opos]=conv4byte2int(ipb.hd+4);
                  opos++;
                  if(opos == nwrite){
                     iwrite=fwrite(obuf,sizeof(int),nwrite,fo);
                     opos=0;
				  }
			  }
		  } 
	   }
exitloop:

	   for(n=0; n<gr.tot_track; n++) i=fclose(mk5[n].fin);

	   /* rest of data write */
	   if(opos >0){
          iwrite=fwrite(obuf,sizeof(int),opos,fo);
	   }
       i=fclose( fo ) ;
       printf("  K5 file (%s) has been created.\n\n",runp.k5name);

       free(mk5);

       if(runp.channel==0 && runp.group==0) goto loop_g;
exit_loop_g:

       tfinish= time(NULL);
       telapsed = difftime(tfinish,tstart);
       printf("Time elapsed (sec) for conversion is %f\n",telapsed);


exitend1:

exitend: 
       return 0;
	   
}
